package net.eduvax.heml;

import java.net.URL;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.Vector;

public class Parser implements Runnable {
    private int _line=1;
    private int _col=1;
    private int _tabSize=4;
	private String _streamName="";
	private InputStream _in;
    private InputStreamReader _reader=null;
	private ParserCallback _handler;
	private char[] _separators={'{','}','%','#','!','+','\\',':'};
	private final static int S_OPEN=0;
	private final static int S_CLOSE=1;
	private final static int S_SEP=2;
	private final static int S_COMMENT=3;
	private final static int S_CDATA=4;
	private final static int S_TBLEXT=5;
	private final static int S_ESC=6;
	private final static int S_INC=7;
	private SInclude _includeState=null;
	public State getState() {
		return _state;
	}
    private void setState(State state) {
        _state=state;
        _handler.stateChanged(state);
    }
	public Parser(InputStream in,ParserCallback handler) {
		_in=in;
		_handler=handler;
        setState(ENTRY);
	}
	public void run() {
        try {
			_reader=new InputStreamReader(_in);
    		int ch=_in.read();
	    	while (ch>=0) {
                char c=(char)ch;
		    	_state.handle(c);
                _col++;
                if (c=='\n') {
                    _line++;
                    _col=1;
                }
                ch=_reader.read();
				if (ch<0&&_includeState!=null) {
					_includeState.popReader();
					ch=_reader.read();
				}
		    }
			_reader.close();
        }
        catch (IOException ex) {
            printErr(""+ex);        
        }
        if (_state!=ENTRY) {
            printErr("Unexpected end of file "+_state);
        }
        _handler.closeDocument();
	}

    public void setParameter(String name,String value) {
        if ("encoding".equals(name)) {
            if (_reader==null) {
                try {
                    _reader=new InputStreamReader(_in,value);
                }
                catch (UnsupportedEncodingException ex) {
                    printErr(""+ex);
                }
            }
            else {
                printErr("Encoding already set ["+_reader.getEncoding()
                    +"]: new "+value+" encoding settings ignored");
            }
        }
        else if ("tab".equals(name)) {
            _tabSize=Integer.parseInt(value);
            // TODO check if value is a valid integer
        }
        else {
            printErr("Unhandled parameter: "+name);
        }
    }


    private void printErr(String msg) {
        System.err.println("Err:"+_streamName+":"+_line+":"+_col+": "+msg);
    }

	private String popAcc() {
		String res=_acc.toString().trim();
		_acc=new StringBuilder();
		return res;
	}

	private int _attrCount=0;
	public void openElement() {
		_attrCount=0;
		_handler.openElement(popAcc());
	}
	private String _attrName=null;
	public void addAttribute() {
		String aName=_attrName;
		if (aName==null) {
			_attrCount++;
			aName="a"+_attrCount;
		}
		_handler.addAttribute(aName,popAcc());
		_attrName=null;
	}
	public void addText() {
		_handler.addText(popAcc());
	}
	
	/** State able to call back specified state */
	public class State {
		private State _retState;
		public void handle(char ch) {
		}
		public State(State retState) {
			_retState=retState;
		}
		public void goBackState() {
			setState(_retState);
		}
		public State getBackState() {
			return _retState;
		}
	}
	/** Entry state */
	private final State ENTRY=new State(null) {
        private String _paramName;
		public void handle(char ch) {
			if (ch==_separators[S_OPEN]) {
					setState(new SElem(this));
			}		
			else if (ch=='=') {
                    _paramName=popAcc();
			}
			else if (ch=='\r') {
			}
            else if (ch=='\n') {
                    setParameter(_paramName,popAcc());
			}		
			else {
                    _acc.append(ch);
			}
		}
	};
	/** current state */
	private State _state;
	/** accumulator */
	private StringBuilder _acc=new StringBuilder();

	/**
	 * Inside attribute name or value.
	 * Or value if anonymous attribute.
	 */
	private class Attr extends State {
		public Attr(State s) {
			super(s);
		}
		public void handle(char ch) {
			if (ch=='=') {
					_attrName=popAcc();
					setState(new AttrV(getBackState()));
			}					
			else if (ch==_separators[S_SEP]) {
					addAttribute();
					setState(new TorA1(getBackState()));
			}
			else if (ch=='\r') {
			}
			else if (ch=='\n') {
					addAttribute();
					setState(new TorA2(getBackState()));
			}
			else if (ch==_separators[S_ESC]) {
					setState(new EscChar(this));
			}
			else if (ch==_separators[S_CLOSE]) {
					addAttribute();
                    _handler.endAttributes();
                    _handler.closeElement();
                    goBackState();
            }
			else {
					_acc.append(ch);
			}
		}
	}
	/**
	 * Inside attribute value.
	 */
	private class AttrV extends State {
		public AttrV(State s) {
			super(s);
		}
		public void handle(char ch) {
			if (ch==_separators[S_SEP]) {
					addAttribute();
					setState(new TorA1(getBackState()));
			}
			else if (ch=='\r') {
			}
			else if (ch=='\n') {
					addAttribute();
					setState(new TorA2(getBackState()));
			}
			else if (ch==_separators[S_ESC]) {
					setState(new EscChar(this));
			}
			else if (ch==_separators[S_CLOSE]) {
					addAttribute();
                    _handler.endAttributes();
                    _handler.closeElement();
                    goBackState();
            }
			else {
					_acc.append(ch);
			}
		}
	}
	/** CData.
	 * Waiting for end CData, ignoring anything else.
	 */
	private class CData extends State {
		public CData(State s) {
			super(s);
		}
		public void handle(char ch) {
			if (ch==_separators[S_CDATA]) {
					setState(new ECData(getBackState()));
			}
			else {
					_acc.append(ch);
			}
		}
	}
	/** End CData.
	 * '}' confirms comment end and go back to before comment state, other leads
	 * back to Comment.
	 */
	private class ECData extends State {
		public ECData(State s) {
			super(s);
		}
		public void handle(char ch) {
			if (ch==_separators[S_CLOSE]) {
					_handler.addCData(popAcc());
					goBackState();
			}
			else {
					_acc.append(_separators[S_CDATA]);
					_acc.append(ch);
					setState(new CData(getBackState()));
			}
		}
	}

	/** Comment.
	 * Waiting for end comment, ignoring anything else.
	 */
	private class Comment extends State {
		public Comment(State s) {
			super(s);
		}
		public void handle(char ch) {
			if (ch==_separators[S_COMMENT]) {
					setState(new EComment(getBackState()));
			}
			else {
					_acc.append(ch);
			}
		}
	}
	/** End Comment.
	 * '}' confirms comment end and go back to before comment state, other leads
	 * back to Comment.
	 */
	private class EComment extends State {
		public EComment(State s) {
			super(s);
		}
		public void handle(char ch) {
			if (ch==_separators[S_CLOSE]) {
					_handler.addComment(popAcc());
					goBackState();
			}
			else {
					_acc.append(_separators[S_COMMENT]);
					_acc.append(ch);
					setState(new Comment(getBackState()));
			}
		}
	}
	/**
	 * Inside element name.
	 */ 
	private class ElemName extends State {
		public ElemName(State s) {
			super(s);
		}
		public void handle(char ch) { 
			if (ch==_separators[S_SEP]) {
						openElement();
						setState(new TorA1(getBackState()));
			}
			else if (ch=='\r') {
			}
			else if (ch=='\n') {
						openElement();
						setState(new TorA2(getBackState()));
			}						
			else if (ch==' ') {
						openElement();
						setState(new TorA3(getBackState()));
			}						
			else if (ch==_separators[S_CLOSE]) {
                    openElement();
	                _handler.endAttributes();
	                _handler.closeElement();
	                goBackState();
			}
			else {
						_acc.append(ch);
			}
		}
	}


	/** Escape char 
	 * anti slash is the escape char.
	 */
	private class EscChar extends State {
		public EscChar(State s) {
			super(s);
		}
		public void handle(char ch) {
			if (ch<='z' && ch>='a') {
				_acc.append(ch-'a');
			}
			else {
				_acc.append(ch);
			}
            goBackState();
		}
	}
	
	/**
	 * count indent chars.
	 */
	private class Indent extends State {
		private int _indent=0;
		private IndentText _prev=null;

		public Indent(State back) {
			super(back);
		}
        public void close() {
            IndentText p=_prev;
			while (p!=null) {
				p.close();
                p=p._prev;
			}
            goBackState();
        }
		public Indent(State back,IndentText prev) {
			super(back);
			_prev=prev;
		}
		public void handle(char ch) {
			if (ch=='\t') {
					_indent+=_tabSize;;
			}
			else if (ch==' ') {
					_indent++;
			}
			else if (ch==_separators[S_OPEN]) {
					if (_prev!=null) {
						_prev.close();
					}
					setState(new SElem(this));
			}
			else if (ch=='-') {
					setState(new IndentText(getBackState(),_indent,_prev,true));
			}
			else if (ch=='\r') {
			}
			else if (ch=='\n') {
					_indent=0;
			}
			else if (ch==_separators[S_CLOSE]) {
					if (_prev!=null) {
						_prev.close();
					}
                    addText();
					_handler.closeElement();
                    goBackState();
			}
			else {
					_acc.append(ch);
					setState(new IndentText(getBackState(),_indent,_prev,false));
			}
		}
	}
	/**
	 * Manage indentation
	 */
	private class IndentText extends State {
		private int _indent=0;
		private IndentText _prev=null;
        private boolean _closed=false;
		private boolean _indentAdded=false;
		private boolean _bullet=false;
		public IndentText(State back,int indent,IndentText prev,boolean bullet) {
			super(back);
			_bullet=bullet;
			_prev=prev;
            _indent=indent;
			IndentText p=_prev;
			while (p!=null && p._indent>=_indent) {
                p.close(); 
				p=p._prev;
			}
			if (_prev!=null&&_prev._indent<_indent) {
				_handler.openIndent();
				_indentAdded=true;
			}
			if (_bullet) {
				_handler.openEnum();
			}
			else {
				_handler.openPara();
			}
		}
        public void close() {
            if (!_closed) {
				if (_bullet) {
	  				_handler.closeEnum();
				}
				else {
					_handler.closePara();
				}
				if (_indentAdded) {
					_handler.closeIndent();
				}
                _closed=true;
            }
        }
		public void handle(char ch) {
			if (ch=='\r') {
			}
			else if (ch=='\n') {
					addText();
					setState(new Indent(getBackState(),this));
			}
			else if (ch==_separators[S_OPEN]) {
                    addText();
					setState(new SElem(this));
			}
			else if (ch==_separators[S_CLOSE]) {
                    addText();
                    close();
					_handler.closeElement();
                    goBackState();
            }
			else if (ch==_separators[S_ESC]) {
					setState(new EscChar(this));
			}
			else {
					_acc.append(ch);
			}
		}
	}
	/**
	 * Inline element text
	 */
	private class InlineText extends State {
		public InlineText(State s) {
			super(s);
		}
        public void close() {
            goBackState();
        }
		public void handle(char ch) {
			if (ch==_separators[S_CLOSE]) {
                    addText();
					_handler.closeElement();
                    goBackState();
			}
			else if (ch==_separators[S_ESC]) {
					setState(new EscChar(this));
			}
			else if (ch==_separators[S_OPEN]) {
                    addText();
					setState(new SElem(this));
			}
			else {
					_acc.append(ch);
			}
		}
	}
	/** 
	 * Opening brace. 
	 * May be start of comment, table or standard element.
	 */ 
	private class SElem extends State {
		public SElem(State s) {
			super(s);
		}
		public void handle(char ch) {
			if (ch==_separators[S_SEP]) {
					setState(new STable(getBackState()));
			}
			else if (ch==_separators[S_INC]) {
					setState(new SInclude(getBackState()));
			}
			else if (ch==_separators[S_COMMENT]) {
					setState(new Comment(getBackState()));
			}
			else {
					_acc.append(ch);
					setState(new ElemName(getBackState()));
			}
		}
	}

	/**
	 * Include start 
	 */
	private class SInclude extends State {
		private InputStreamReader _pushedReader=null;
		private SInclude _pushedInclude=null;
		private String _pushedStreamName="";
		private int _pushedCol=1;
		private int _pushedLine=1;
		public SInclude(State s) {
			super(s);
		}
		public void popReader() {
			if (_pushedReader!=null) {
				try {
					_reader.close();
				}
				catch (IOException ex) {
					printErr(ex.getMessage());
				}
				_reader=_pushedReader;
				_includeState=_pushedInclude;
				_streamName=_pushedStreamName;
				_line=_pushedLine;
				_col=_pushedCol;
			}
		}
		public void handle(char ch) {
			if (ch==_separators[S_CLOSE]) {
				String urlStr=popAcc();
				try {
					InputStream input=null;
					InputStreamReader includeReader=null;
					File f=new File(urlStr);
					if (f.exists()) {
						input=new FileInputStream(f);
					}
					else {
						URL url=new URL(urlStr);
						input=url.openStream();
					}
					includeReader=new InputStreamReader(input,
                            _reader.getEncoding());
					_pushedLine=_line;
					_line=1;
					_pushedCol=_col;
					_col=1;
					_pushedStreamName=_streamName;
					_streamName=urlStr;
					_pushedReader=_reader;
					_reader=includeReader;
					_pushedInclude=_includeState;
					_includeState=this;
				}
				catch (Exception ex) {
					printErr("Can't include "+urlStr+": "+ex.getMessage());
				}
				goBackState();
			}
			else {
				_acc.append(ch);
			}
		}
	}
    /**
     * Table start : row name.
     */
    private class STable extends State {
        public STable(State s) {
            super(s);
        }
        public void handle(char ch) {
            if (ch==_separators[S_SEP]) {
                    setState(new TableCell(getBackState(),popAcc()));
			}
			else {
                    _acc.append(ch);
            }
        }
    }
    /**
     * Table cell decl.
     */
    private class TableCell extends State {
        private String _rowName;
        private Vector<String> _cellNames=new Vector<String>();
        public TableCell(State s,String rowName) {
            super(s);
            _rowName=rowName;
        }
        public void handle(char ch) {
            if (ch=='\r') {
			}
            else if (ch=='\n') {
                    _cellNames.add(popAcc());
					int style=0;
					if (_rowName.charAt(0)==_separators[S_TBLEXT]) {
						style=1;
						_rowName=_rowName.substring(1);
					}
					_handler.openTable(style,_rowName,_cellNames);
                    setState(new RowContent(getBackState()));
			}
            else if (ch==_separators[S_SEP]) {
                    _cellNames.add(popAcc());
			}
            else {
                    _acc.append(ch);
            }
        }
    }

    /**
     * Table row content
     */
    private class RowContent extends State {
        private Vector<String> _cellValues=new Vector<String>();
        public RowContent(State s) {
            super(s);
        }
		public void close() {
			if (_cellValues.size()>0) {
				_handler.addRow(_cellValues);
			}	
			_cellValues=new Vector<String>();
		}
		public void addField() {
			_cellValues.add(popAcc());
		}
        public void handle(char ch) {
			if (ch=='\r') {
			}
            else if (ch=='\n') {
                    addField();
					close();
			}
            else if (ch==_separators[S_SEP]) {
                    addField();
			}
            else if (ch==_separators[S_CLOSE]) {
                    String f=popAcc();
                    if (f.length()>0 || _cellValues.size()>0) {
			            _cellValues.add(f);
                    }
                    close();
                    goBackState();
			}
            else {
                    _acc.append(ch);
            }
        }
    }

	/**
	 * Text or attribute case 1.
	 * % encountered, follwoing may be attribute, element end or inline text.
	 */
	private class TorA1 extends State {
		public TorA1(State s) {
			super(s);
		}
		public void handle(char ch) {
            if (ch==_separators[S_SEP]) {
					_handler.endAttributes();
					setState(new InlineText(getBackState()));
			}
            else if (ch=='\r') {
			}
            else if (ch=='\n') {
                    setState(new TorA2(getBackState()));
			}
            else if (ch==_separators[S_CLOSE]) {
					_handler.endAttributes();
					_handler.closeElement();
					goBackState();
			}
            else {
					_acc.append(ch);
					setState(new Attr(getBackState()));
			}
		}
	}
	/**
	 * Text or attribute case 2.
	 * % encountered, follwoing may be:
	 * - attribute indented in new line, 
	 * - structured text.  
	 */
	private class TorA2 extends State {
        private boolean _endAttribute=false;
		public TorA2(State s) {
			super(s);
		}
		private int _indent=0;
        private void endAttributes() {
            if (!_endAttribute) {
                _endAttribute=true;
                _handler.endAttributes();
            }
        }
		public void handle(char ch) {
            if (ch=='\r'||ch=='\n') {
			}
            else if (ch=='\t') {
					_indent+=_tabSize;
			}
            else if (ch==' ') {
					_indent++;
			}
            else if (ch==_separators[S_SEP]) {
					setState(new TorA1(getBackState()));
			}
            else if (ch=='-') {
					endAttributes();
					setState(new IndentText(getBackState(),_indent,null,true));
			}
            else if (ch==_separators[S_OPEN]) {
                    endAttributes();
                    setState(new SElem(new Indent(getBackState()))); 
			}
            else if (ch==_separators[S_CLOSE]) {
					_handler.endAttributes();
					_handler.closeElement();
					goBackState();
			}
            else {
					endAttributes();
                    _acc.append(ch);
					setState(new IndentText(getBackState(),_indent,null,false));
			}
		}
	}
	/**
	 * Text or attribute case 3.
	 * space encountered, element name is known, following may be inline text 
	 * or attributes.
	 */
	private class TorA3 extends State {
		public TorA3(State s) {
			super(s);
		}
		public void handle(char ch) {
            if (ch==_separators[S_SEP]) {
					setState(new TorA1(getBackState()));
			}
            else if (ch=='\t'||ch==' ') {
			}
            else if (ch==_separators[S_CLOSE]) {
					_handler.endAttributes();
					_handler.closeElement();
					goBackState();
			}
            else {
                    _handler.endAttributes();
					_acc.append(ch);
					setState(new InlineText(getBackState()));
			}
		}
	}
}
